from collections import defaultdict
import re
from xml.etree import cElementTree as ET
import os
import time
import json
import logging
import urllib.request
import urllib.parse
import urllib.error
import jetpack.util

logger = logging.getLogger()


def lockdown_userdata(userdata_file):
    '''
    lock down the userdata file so that it is not world readable
    Only necessary on Linux
    '''
    if not jetpack.util.is_windows():
        os.chmod(userdata_file, 0o640)


def cache_userdata(refresh_cache=False):
    """ Cache the userdata to disk and return the JSON representation """
    cached_file = os.path.join(jetpack.util.jetpack_home(), "config", "user-data.json")
    # Check to see if we've already cached the userdata on disk, if yes use that
    # We want to update SAS token every 30 minutes to have some extra time until expiration (it expires after 1 hour)
    if not refresh_cache and os.path.exists(cached_file) and os.path.getsize(cached_file) > 0:
        with open(cached_file) as f:
            json_data = f.read()
        data = json.loads(json_data)
        return data
    else:
        ud = Userdata()
        json_data = ud.to_json()

    # Ensure that the bootstrap directory exists before writing to it
    if not os.path.isdir(os.path.dirname(cached_file)):
        os.makedirs(os.path.dirname(cached_file))

    with open(cached_file, 'w') as f:
        f.write(str(json_data))

    lockdown_userdata(cached_file)

    return json.loads(json_data)


def _get_customdata(refresh_cache=False):
    """ Parses the embedded customdata from the legacy XML """
    ud = cache_userdata(refresh_cache)
    return json.loads(ud.get('user_data', {}).get('customdata', '{}'))


def get_config():
    """ Gets the user configuration from userdata """
    customdata = _get_customdata()
    return customdata.get('configuration', {})


def get_settings():
    """ Gets the node settings from userdata """
    customdata = _get_customdata()
    return customdata.get('settings', {})


def get_lockers(refresh_cache=False):
    """ Gets the lockers from userdata """
    customdata = _get_customdata(refresh_cache)
    return customdata.get('lockers', {})


def etree_to_dict(t):
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update((k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d


class Userdata():
    def __init__(self):
        self.userdata_hash = "" # Userdata as a python dict
        self.userdata_xml = ""  # Userdata as an XML string

    def fetch(self):
        """ Fetch the userdata from CycleCloud """

        connection = jetpack.util.get_connection_config()

        cluster_name = connection['cluster']
        instance_id = jetpack.util.get_instance_id()

        url = "/clusterlink/userdata/%s" % urllib.parse.quote(instance_id)

        params = {
            'cluster': cluster_name
        }

        logger.info("Getting node configuration from %s" % url)
        r = jetpack.util.query_cyclecloud(url, params=params)

        if r.status != 200:
            logger.exception("An error occurred getting node configuration %s: %s (%s)" % (url, r.reason, r.status))
            raise Exception(r.reason)
        if "application/xml" not in r.getheader('Content-Type').lower():
            raise Exception("Configuration received from %s was not application/xml" % url)

        self.userdata_xml = r.read().decode()

    def to_xml(self):
        if not self.userdata_xml:
            self.fetch()
        return self.userdata_xml

    def to_dict(self):
        """ convert the xml to dictionary. """
        if not self.userdata_xml:
            self.to_xml()
        if not self.userdata_hash:
            # remove the namespace
            xml = re.sub(' xmlns="[^"]+"', '', self.userdata_xml, count=1)
            e = ET.XML(xml)
            self.userdata_hash = etree_to_dict(e)
        return self.userdata_hash

    def to_json(self):
        """ convert the dictionary to valid json.
            NOTE: we have to add quite a bit of additional helper data to the json.
            NOTE: we also are renaming the top-level key. """
        if not self.userdata_hash:
            self.to_dict()

        # We need to convert the top-level user-data key to user_data
        self.userdata_hash['user_data'] = self.userdata_hash.pop('user-data')

        json_data = json.dumps(self.userdata_hash, indent=2, sort_keys=True)
        logger.debug("Converted raw userdata to json format.")
        return json_data
