import os
from jetpack import util
from jetpack import messaging


class SendError(Exception):
    pass


class LogError(Exception):
    pass


def filter_metadata(metadata, required_fields):
    return {k: metadata[k] for k in required_fields}


def _get_priority(level):
    if level == 'info':
        log_priority = 'medium'
    elif level == 'warn':
        log_priority = 'medium'
    elif level == 'error':
        log_priority = 'high'
    else:
        raise LogError("Invalid log level")
    return log_priority


# this is here because send_internal_message attempts to read jetpack.ini file which is not fully ready before the converge
def prepare_messaging_config(cluster_session_id, cluster_name, instance_id):
    config = {}
    config['identity'] = {}
    config['identity']['cluster_session_id'] = cluster_session_id
    config['identity']['cluster_name'] = cluster_name
    config['identity']['instance_id'] = instance_id
    config['cycle_server'] = {}
    return config


def send_internal_message_before_node_is_ready(message_data, message_type, cluster_session_id):
    '''
    This is a wrapper for send_internal_message for the case when the converge is not finished
    parameters:
      message_data - this is a python dictionary
      message_type - type of message, examples are test, log, installation
    '''
    if not isinstance(message_data, dict):
        raise SendError("message_data parameter must be a dictionary")

    # since we store the logs on our side too, unnecessary info should be trimmed away
    filtered_metadata_fields = {"location", "offer", "osType", "provider", "version", "vmScaleSetName", "vmSize"}
    tags_list = message_data["vm_metadata"]["tagsList"]
    cluster_name = None
    resource_id = message_data["vm_metadata"]["resourceId"]
    vm_id = message_data["vm_metadata"]["vmId"]
    for tag in tags_list:
        if tag["name"] == "ClusterName":
            cluster_name = tag["value"]
    # tags are no longer needed, so we can filter them out
    message_data["vm_metadata"] = filter_metadata(message_data["vm_metadata"], filtered_metadata_fields)
    # there is no real instance_id, so might as well set it to something useful
    instance_id = util.get_instance_id()
    # ideally this should be in the message_obj, but that would require changing the system message handler
    message_data["resource_id"] = resource_id
    message_data["vm_id"] = vm_id
    
    # create a config for send_internal_message to use
    config = prepare_messaging_config(cluster_session_id, cluster_name, instance_id)
    
    messaging.send_internal_message(message_data, message_type, config)


def report_error(test_data, level="error", priority=None, cluster_session_id=None):
    '''
    Reports a NodeHealthChecks error back to CycleCloud. test_data must include the exit code, the name of the test and a vm metadata
    '''
    if level not in ["info", "warn", "error"]:
        raise LogError("Invalid level: %s" % level)

    priority = priority or _get_priority(level)

    try:
        vm_metadata = test_data['vm_metadata']
        vm_id = vm_metadata['vmId']
        
        if level == "error":
            message = f"Node {vm_id} failed healthcheck {os.path.basename(test_data['test_path'])}"
        else:
            message = f"HealthChecks didn't run on node {vm_id}"
        
        message_data = {
            'level': "info",  # all nodehealthchecks errors are reported to our AppInsights at info severity level
            'message': message,
            'priority': priority,
            'exit_code': test_data['exit_code'],
            'vm_metadata': vm_metadata,
            'test_path': test_data['test_path'],
            'test_output': test_data['message']
        }
        
    except KeyError as exc:
        raise LogError("Invalid healthcheck report") from exc
    send_internal_message_before_node_is_ready(message_data, "node_health_checks", cluster_session_id)
