import unittest
from unittest import mock
from jetpack import converge
from jetpack import cluster_init
import jetpack.status
import jetpack.util
import os
import pytest  # type: ignore


@pytest.mark.parametrize("distro_id,coerced", [
    ['rocky', 'rhel'],
    ['almalinux', 'rhel'],
    ['RoCKy', 'rhel'],
    ['AlmaLinux', 'rhel'],
    ['Ubuntu', ''],
    ['CentOS', ''],
    ['SLE-hpc', 'suse'],
    ['sle_HPC', 'suse']
])
def test_coerce_platform_family(distro_id, coerced):
    with mock.patch('distro.id', return_value=distro_id):
        config = converge._coerce_platform_family({})
        if coerced:
            assert config['platform_family'] == coerced
        else:
            assert 'platform_family' not in config

class TestConverge(unittest.TestCase):
    def setUp(self):
        self.linux_bootstrap = "/opt/cycle/jetpack/system/bootstrap"
        self.linux_logs = "/opt/cycle/jetpack/logs"
        self.win_bootstrap = "c:\\cycle\\jetpack\\system\\bootstrap"
        self.win_logs = "c:\\cycle\\jetpack\\logs"

    @mock.patch('os.path.isfile', return_value=False)
    @mock.patch('jetpack.util.is_linux', return_value=True)
    @mock.patch('jetpack.config.get', return_value='http://1.2.3.4')
    def test_setup_proxy(self, mock_get, mock_is_linux, mock_is_file):
        with mock.patch.dict('os.environ', {}):
            with mock.patch('jetpack.converge.open', mock.mock_open(), create=True) as mock_open:
                with mock.patch('os.chmod') as mock_chmod:
                    converge._setup_proxy()
                    mock_get.assert_called_once_with('http_proxy.url', '')
                    mock_is_file.assert_called_once_with("/etc/profile.d/proxy.sh")

                    mock_open.return_value.__enter__.return_value.write.assert_has_calls([
                        mock.call("# Written by CycleCloud\n"),
                        mock.call("export http_proxy=http://1.2.3.4\n"),
                        mock.call("export https_proxy=http://1.2.3.4\n"),
                        mock.call("export no_proxy=169.254.169.254  # special rule exempting metadata URL from proxy\n")
                    ])

                    mock_chmod.assert_called_once_with("/etc/profile.d/proxy.sh", 0o644)

                    self.assertEqual('http://1.2.3.4', os.environ['http_proxy'])
                    self.assertEqual('http://1.2.3.4', os.environ['https_proxy'])
                    self.assertEqual('169.254.169.254', os.environ['no_proxy'])

    def test_read_timestamps(self):
        with mock.patch('os.path.isfile', return_value=False) as mock_isfile:
            timestamps = converge._read_timestamps()
            mock_isfile.assert_called_once_with("/opt/cycle/jetpack/config/timestamps.json")
            self.assertEqual({}, timestamps)

        with mock.patch('os.path.isfile', return_value=True) as mock_isfile:
            with mock.patch('jetpack.converge.open', mock.mock_open(), create=True) as mock_open:
                mock_open.return_value.__enter__.return_value.read.return_value = """
                {"timestamp": "oldtimedata", "othertimestamp": "data"}
                """
                timestamps = converge._read_timestamps()
                self.assertEqual({
                    "othertimestamp": "data",
                    "timestamp": "oldtimedata"
                }, timestamps)

    @mock.patch('jetpack.converge.socket')
    def test_send_installation_status(self, socket):
        socket.gethostname.return_value = 'test-hostname'
        socket.getfqdn.return_value = 'test-fqdn'
        with mock.patch('jetpack.messaging.send_internal_message') as mock_send:
            with mock.patch('jetpack.converge._read_timestamps', return_value={"timestamp": "in-da-past"}) as mock_read_timestamps:
                converge._send_installation_status("success")

                mock_read_timestamps.assert_called_once_with()
                mock_send.assert_called_once_with(
                    {'status': 'success', 'timestamp': 'in-da-past', 'update_status': False,
                     'hostname': 'test-hostname', 'fqdn': 'test-fqdn'}, 'installation')
                
    @mock.patch('jetpack.cluster_init.get_specs', return_value=[])
    def test_update_runlist(self, mock_get_specs):
        # Empty runlist
        self.assertEqual({
            'run_list': ["recipe[cyclecloud]", "recipe[cluster_init]"]
        }, converge._update_runlist({
            'run_list': []
        }))

        # Runlist with cyclecloud already in it and a test recipe
        self.assertEqual({
            'run_list': ["recipe[cyclecloud]", "recipe[test]", "recipe[cluster_init]"]
        }, converge._update_runlist({
            'run_list': ["recipe[cyclecloud]", "recipe[test]"]
        }))

        # Runlist with runlists from specs (user defined)
        mock_get_specs.return_value = [
            cluster_init.ClusterInitSpec({
                'project': 'test',
                'project_spec': 'default',
                'version': '1.0.0',
                'location': 'file:///foo/bar/',
                'run_list': ['recipe[a]', 'recipe[b]']
            })
        ]

        self.assertEqual({
            'run_list': ['recipe[cyclecloud]', 'recipe[a]', 'recipe[b]', 'recipe[cluster_init]'],
        }, converge._update_runlist({
            'run_list': []
        }))    

    @mock.patch('jetpack.cluster_init.process_chef')
    @mock.patch('jetpack.util.is_linux')
    def test_sync_chef_repos(self, mock_is_linux, mock_process_chef):
        converge._sync_chef_repos(bootstrap_path="/tmp/bootstrap", log_path="/tmp/logs")
        mock_process_chef.assert_called_once()

    @mock.patch('jetpack.util.is_linux')
    def test_run_chef_solo(self, mock_is_linux):
        mock_is_linux.return_value = True
        with mock.patch('jetpack.converge.subprocess.Popen') as mock_popen:
            # check the command called for linux
            mock_popen.return_value = mock.MagicMock(returncode=0)
            status = converge._run_chef_solo(log_path="/tmp/logs", flags="")
            self.assertEqual('success', status)
            mock_popen.assert_called_once_with("chef-solo -c /opt/cycle/jetpack/system/chef/solo.rb -L /tmp/logs/chef-client.log ", shell=True)

            # check that the correct cmd is called for windows
            mock_popen.reset_mock()
            mock_popen.return_value = mock.MagicMock(returncode=0)
            mock_is_linux.return_value = False
            status = converge._run_chef_solo(log_path="C:\\temp\\logs", flags="--test")
            self.assertEqual('success', status)            
            mock_popen.assert_called_once_with("chef-solo.bat -c C:\\cycle\\jetpack\\system\\chef\\solo.rb -L C:\\temp\\logs\\chef-client.log --test", shell=True)

            # assert that the command fails then a non-zero exit code is returned
            mock_popen.reset_mock()
            mock_popen.return_value = mock.MagicMock(returncode=5)
            self.assertRaises(converge.ConvergeError, converge._run_chef_solo, log_path="", flags="")


    @mock.patch('jetpack.util.is_linux')
    @mock.patch('jetpack.util.is_windows') 
    def test_run_chef_reboot(self, mock_is_windows, mock_is_linux):
        mock_is_linux.return_value = True
        mock_is_windows.return_value = False
        with mock.patch('jetpack.converge.subprocess.Popen') as mock_popen:
            # linux should fail when it get the 'reboot' error code
            mock_popen.return_value = mock.MagicMock(returncode=35)
            self.assertRaises(converge.ConvergeError, converge._run_chef_solo, log_path="", flags="")
            # windows should not fail and indicate a reboot is happening
            mock_popen.reset_mock()
            mock_popen.return_value = mock.MagicMock(returncode=35)
            mock_is_linux.return_value = False
            mock_is_windows.return_value = True
            status = converge._run_chef_solo(log_path="C:\\temp\\logs", flags="--test")
            self.assertEqual('rebooting', status)



    @mock.patch('jetpack.util.get_current_config')
    @mock.patch('jetpack.converge.userdata.get_settings', return_value={})
    @mock.patch('jetpack.converge.userdata.get_lockers', return_value={})
    @mock.patch('datetime.datetime')
    @mock.patch('jetpack.util.write_timestamps')
    @mock.patch('jetpack.status.set')
    @mock.patch('jetpack.lockers.update')
    @mock.patch('jetpack.settings.update')
    @mock.patch('jetpack.converge.os.remove')
    @mock.patch('jetpack.converge.os.path.isfile', return_value=False)
    @mock.patch('jetpack.config.update')
    @mock.patch('jetpack.cluster_init.get_specs', return_value=[])
    @mock.patch('jetpack.converge.userdata.get_config', return_value={})
    @mock.patch('jetpack.util.parse_config')
    @mock.patch('jetpack.config.get', return_value=False)
    @mock.patch('jetpack.converge._write_configuration')
    @mock.patch('jetpack.util.get_boottime')
    @mock.patch('jetpack.util.clear_configuration_status')
    @mock.patch('jetpack.util.rebuild_rpmdb', return_value=False)
    @mock.patch('jetpack.util.write_configuration_exception')
    @mock.patch('jetpack.util.read_configuration_status')
    def test_linux_fail_get_specs(self, mock_read_configuration_status, mock_write_configuration_exception, mock_rebuild_rpmdb, mock_clear_configuration_status, mock_get_boottime, mock_write_configuration, mock_config_get, mock_parse_config, mock_get_config, mock_get_specs, mock_config_update, mock_isfile, mock_remove, mock_settings_update, mock_lockers_update, mock_status_set,  mock_write_timestamps, mock_datetime, mock_get_lockers, mock_get_settings, mock_get_current_config):
        
        # Set the current time, and the initial start time to the same value signaling "initial converge"
        timestamp = '2001-01-01:11:11:11'
        mock_datetime.now.return_value.isoformat.return_value = timestamp
        mock_write_timestamps.return_value = {
            'converge_initial_start_timestamp': timestamp,
            'converge_initial_finish_timestamp': '2002-02-02:22:22:22'
        }
        mock_parse_config.return_value = {
            "identity": {
                "cluster_name": "c1",
                "instance_id": "i-123"
            },
            "amqp": {}
        }
        mock_write_configuration.return_value = mock_parse_config.return_value
        status_mock = mock.Mock()
        status_mock.return_value = mock.Mock(), mock.Mock()
        mock_read_configuration_status.return_value = { 
            'message': "Error configuring software",
            'description': 'description',
            'detail': 'detail',
            'recommendation': 'recommendation',
            'help_link': 'help_link' 
        }
        mock_get_specs.side_effect=jetpack.util.UserError("The source file https://fake_locker/fake_spec cannot be verified!")
        with mock.patch.multiple("jetpack.converge", _run_chef_solo=mock.DEFAULT,
                                                     _sync_chef_repos=mock.DEFAULT,                                                   
                                                     _send_installation_status=mock.DEFAULT,
                                                     _coerce_platform_family=mock.DEFAULT) as funcs:
            with mock.patch('jetpack.util.already_running', return_value=False):
                with mock.patch('jetpack.amqp.send', return_value=None):
                    # Test chef-solo
                    funcs['_run_chef_solo'].return_value = 'success'

                    # Verify that at the end of a converge we blow away the pid file
                    mock_isfile.return_value = True
                    with mock.patch('jetpack.config.get', return_value=False):
                        with self.assertRaises(jetpack.util.UserError):
                            converge.execute(debug=False, mode='install')

                    mock_get_specs.assert_called_once()
                    mock_write_configuration_exception.assert_called_once()

                    # Update status should have been called twice, once at the start and once upon error
                    mock_status_set.assert_has_calls([
                        mock.call(jetpack.status.JetpackResource.SoftwareConfiguration,
                                  status=jetpack.status.JetpackStatus.NoStatus,
                                  message="Configuring software",
                                  description="Configuring software on VM",
                                  active=True,
                                  start_time='2001-01-01:11:11:11'),
                        mock.call(name=jetpack.status.JetpackResource.SoftwareConfiguration, 
                                  status=jetpack.status.JetpackStatus.Error, 
                                  message="Error configuring software", 
                                  description="description",
                                  detail='detail', 
                                  recommendation="recommendation",
                                  help_link="help_link",
                                  active=True, 
                                  start_time='2001-01-01:11:11:11')
                    ])


    @mock.patch('jetpack.util.get_current_config')
    @mock.patch('jetpack.converge.userdata.get_settings', return_value={})
    @mock.patch('jetpack.converge.userdata.get_lockers', return_value={})
    @mock.patch('datetime.datetime')
    @mock.patch('jetpack.util.write_timestamps')
    @mock.patch('jetpack.status.set')
    @mock.patch('jetpack.lockers.update')
    @mock.patch('jetpack.settings.update')
    @mock.patch('jetpack.converge.os.remove')
    @mock.patch('jetpack.converge.os.path.isfile', return_value=False)
    @mock.patch('jetpack.config.update')
    @mock.patch('jetpack.cluster_init.get_specs', return_value=[])
    @mock.patch('jetpack.converge.userdata.get_config', return_value={})
    @mock.patch('jetpack.util.parse_config')
    @mock.patch('jetpack.config.get', return_value=False)
    @mock.patch('jetpack.converge._write_configuration')
    @mock.patch('jetpack.util.get_boottime')
    @mock.patch('jetpack.util.clear_configuration_status')
    @mock.patch('jetpack.status.update')
    def test_linux_converge(self, mock_status_update, mock_clear_configuration_status, mock_get_boottime, mock_write_configuration, mock_config_get, mock_parse_config, mock_get_config, mock_get_specs, mock_config_update, mock_isfile, mock_remove, mock_settings_update, mock_lockers_update, mock_status_set,  mock_write_timestamps, mock_datetime, mock_get_lockers, mock_get_settings, mock_get_current_config):

        # Set the current time, and the initial start time to the same value signaling "initial converge"
        timestamp = '2001-01-01:11:11:11'
        mock_datetime.now.return_value.isoformat.return_value = timestamp
        mock_write_timestamps.return_value = {
            'converge_initial_start_timestamp': timestamp,
            'converge_initial_finish_timestamp': '2002-02-02:22:22:22'
        }
        mock_parse_config.return_value = {
            "identity": {
                "cluster_name": "c1",
                "instance_id": "i-123"
            },
            "amqp": {}
        }
        mock_write_configuration.return_value = mock_parse_config.return_value
        status_mock = mock.Mock()
        status_mock.return_value = mock.Mock(), mock.Mock()
        with mock.patch.multiple("jetpack.converge", _run_chef_solo=mock.DEFAULT,
                                                     _sync_chef_repos=mock.DEFAULT,                                                   
                                                     _send_installation_status=mock.DEFAULT,
                                                     _coerce_platform_family=mock.DEFAULT) as funcs:
            with mock.patch('jetpack.util.already_running', return_value=False):
                with mock.patch('jetpack.amqp.send', return_value=None):
                    # Test chef-solo
                    funcs['_run_chef_solo'].return_value = 'success'

                    # Verify that at the end of a converge we blow away the pid file
                    mock_isfile.return_value = True
                    with mock.patch('jetpack.config.get', return_value=False):
                        converge.execute(debug=False, mode='install')

                    mock_remove.assert_called_once_with("/opt/cycle/jetpack/system/run/converge.pid")
                    self.assertEqual(2, mock_write_configuration.call_count)
                    mock_get_config.assert_called_once()
                    mock_lockers_update.assert_called_once()
                    mock_settings_update.assert_called_once()
                    funcs['_sync_chef_repos'].assert_called_once_with(self.linux_bootstrap, self.linux_logs)
                    funcs['_run_chef_solo'].assert_called_once_with(self.linux_logs, '')

                    # Update status should have been called twice, once at the start and once at the end
                    mock_status_set.assert_has_calls([
                        mock.call(jetpack.status.JetpackResource.SoftwareConfiguration,
                                  status=jetpack.status.JetpackStatus.NoStatus,
                                  message="Configuring software",
                                  description="Configuring software on VM",
                                  active=True,
                                  start_time='2001-01-01:11:11:11'),
                        mock.call(name=jetpack.status.JetpackResource.SoftwareConfiguration, 
                                  status=jetpack.status.JetpackStatus.Ok, 
                                  message="Completed software configuration", 
                                  description="Completed software configuration", 
                                  active=False, 
                                  start_time='2001-01-01:11:11:11',
                                  end_time='2002-02-02:22:22:22')
                    ])

                    funcs['_send_installation_status'].assert_has_calls([
                        mock.call('started'),
                        mock.call('success')
                    ])

                    # Try again with no-sync enabled
                    funcs['_sync_chef_repos'].reset_mock()
                    converge.execute(no_sync=True)
                    self.assertFalse(funcs['_sync_chef_repos'].called)
    
    @mock.patch('jetpack.util.get_current_config')
    @mock.patch('jetpack.converge.userdata.get_settings', return_value={})
    @mock.patch('jetpack.converge.userdata.get_lockers', return_value={})
    @mock.patch('jetpack.util.write_timestamps')
    @mock.patch('jetpack.status.set')
    @mock.patch('jetpack.lockers.update')
    @mock.patch('jetpack.settings.update')
    @mock.patch('jetpack.converge.os.remove')
    @mock.patch('jetpack.converge.os.path.isfile', return_value=False)
    @mock.patch('jetpack.cluster_init.get_specs')
    @mock.patch('jetpack.config.update')
    @mock.patch('jetpack.converge.userdata.get_config', return_value={})
    @mock.patch('jetpack.util.parse_config')
    @mock.patch('jetpack.config.get', return_value=False)
    @mock.patch('jetpack.converge._write_configuration')
    @mock.patch('jetpack.util.get_boottime')
    @mock.patch('jetpack.util.clear_configuration_status')
    def test_windows_converge(self, mock_clear_configuration_status, mock_get_boottime, mock_write_configuration, mock_config_get, mock_parse_config, mock_get_config, mock_config_update, mock_get_specs, mock_isfile, mock_remove, mock_settings_update, mock_lockers_update, mock_status_set, mock_write_timestamps, mock_get_lockers, mock_get_settings, mock_get_current_config):
        mock_parse_config.return_value = {
            "identity": {
                "cluster_name": "c1",
                "instance_id": "i-123"
            },
            "amqp": {}
        }
        mock_write_configuration.return_value = mock_parse_config.return_value
        status_mock = mock.Mock()
        status_mock.return_value = mock.Mock(), mock.Mock()

        with mock.patch.multiple("jetpack.converge", _run_chef_solo=mock.DEFAULT,
                                                     _sync_chef_repos=mock.DEFAULT,
                                                     _send_installation_status=mock.DEFAULT,
                                                     _coerce_platform_family=mock.DEFAULT) as funcs:
            with mock.patch('jetpack.util.already_running', return_value=False):
                with mock.patch('platform.system', mock.MagicMock(return_value="Windows")):
                    with mock.patch('jetpack.amqp.send', return_value=None):
                        # Test chef-solo
                        converge.execute(debug=False, mode='install')

                        mock_get_config.assert_called_once()
                        self.assertEqual(2, mock_write_configuration.call_count)
                        funcs['_sync_chef_repos'].assert_called_once_with(self.win_bootstrap, self.win_logs)
                        funcs['_run_chef_solo'].assert_called_once_with(self.win_logs, '')

                        # Verify that we checked to see if the pid file existed (did not) and then didn't attempt to remove the non-existing file
                        mock_isfile.assert_called_once_with("c:\\cycle\\jetpack\\system\\run\\converge.pid")
                        self.assertFalse(mock_remove.called)

    def test_already_running(self):
        # Raises ConvergeError since it's already runing
        with mock.patch('platform.system', mock.MagicMock(return_value="Linux")):
            with mock.patch('jetpack.util.already_running', return_value=True):
                self.assertRaises(converge.ConvergeError, converge.execute)
