import os
import tempfile
import unittest
from unittest import mock
from jetpack.converge import userdata
from xml.etree import cElementTree as ET


class TestUserdata(unittest.TestCase):
    def setUp(self):
        self.ud = userdata.Userdata()

    def test_etree_to_dict(self):
        test_data = """<?xml version="1.0"?>
<data>
    <place name="Andorra">
        <index>99</index>
    </place>
    <place name="San Marino">
        <index>2</index>
    </place>
    <place name="Lesotho">
        <index>5</index>
    </place>
</data>"""
        t = ET.fromstring(test_data)
        self.assertEqual({
            "data": {
                "place": [
                    {"name": "Andorra", "index": '99'},
                    {"name": "San Marino", "index": '2'},
                    {"name": "Lesotho", "index": '5'}
                ]
            }
        }, userdata.etree_to_dict(t))

    @mock.patch('jetpack.util.query_cyclecloud')
    @mock.patch('jetpack.util.get_connection_config')
    @mock.patch('jetpack.util.get_instance_id', return_value='instance-1234')
    def test_fetch(self, get_instance_id, get_connection_config, query_cyclecloud):
        get_connection_config.return_value = {
            'cluster': 'test-cluster'
        }
        response = mock.Mock()
        response.status = 200
        response.getheader.return_value = 'application/xml'
        response.read.return_value = b'<sample><xml></xml></sample>'
        query_cyclecloud.return_value = response

        self.ud.fetch()
        query_cyclecloud.assert_called_once_with('/clusterlink/userdata/instance-1234', params={'cluster': 'test-cluster'})
        self.assertEqual("<sample><xml></xml></sample>", self.ud.userdata_xml)

        # Non-200 response should trigger an error
        response.status = 500
        with self.assertRaises(Exception):
            self.ud.fetch()

        # Non text/xml content-type
        response.status = 200
        response.getheader.return_value = 'text/html'
        with self.assertRaises(Exception):
            self.ud.fetch()

    def test_to_xml(self):
        with mock.patch.object(self.ud, 'fetch') as mock_fetch:
            def side_effect():
                self.ud.userdata_xml = "testing"
            mock_fetch.return_value = None
            mock_fetch.side_effect = side_effect
            self.assertEqual("testing", self.ud.to_xml())

            mock_fetch.reset_mock()
            self.ud.userdata_xml = "already set"
            self.assertEqual("already set", self.ud.to_xml())

    def test_to_json(self):
        self.ud.userdata_hash = {
            'user-data': {
                'sup': 'dawg'
            }
        }
        import json
        expected = json.dumps({
            "user_data": {
                "sup": "dawg"
            }
        }, indent=2)
        self.assertEqual(expected, self.ud.to_json())

    @mock.patch('jetpack.converge.userdata._get_customdata')
    def test_get_config(self, _get_customdata):
        _get_customdata.return_value = {
            'configuration': {
                'user-defined-config': 'test-value'
            }
        }
        self.assertEqual('test-value', userdata.get_config()['user-defined-config'])

    @mock.patch('jetpack.converge.userdata._get_customdata')
    def test_get_settings(self, _get_customdata):
        _get_customdata.return_value = {
            'settings': {
                'locker': 'test-locker'
            }
        }
        self.assertEqual('test-locker', userdata.get_settings()['locker'])

    @mock.patch('jetpack.converge.userdata._get_customdata')
    def test_get_lockers(self, _get_customdata):
        _get_customdata.return_value = {
            'lockers': {
                'my-locker': 'locker-data'
            }
        }
        self.assertEqual('locker-data', userdata.get_lockers()['my-locker'])